import json
import os

from pscr_common.utils import _logger

from org_mapper.timestream_util import build_query, execute_query


class CacheTSMetrics:
    metric_timeseries_cache = None

    def __init__(self, look_back_metric_points, config, identifier_column="org_key"):
        self.database = config.get("timestream", "db_name")
        self.table = config.get("timestream", "table_name")
        self.env = config.get("env", "id")
        self.metric_period = config.get("timestream", "metric_period")
        self.look_back_metric_points = look_back_metric_points
        self.top_x_points = config.getint("metrics", "top_x_points")
        self.identifier_column = identifier_column
        self.mode = config.get("service", "mode")

    def load_metric_timeseries_to_cache(
        self, paginator, start_time, end_time, metric_list
    ):
        CacheTSMetrics.metric_timeseries_cache = {}
        metric_timeseries_data = []

        for metrics in metric_list:
            metric_timeseries_data.append({metrics[0]: {}})

        query = build_query(
            self.identifier_column,
            self.metric_period,
            self.env,
            self.database,
            self.table,
            start_time,
            end_time,
            metric_list,
        )

        execute_query(paginator, query, metric_timeseries_data)

        for metric_map in metric_timeseries_data:
            metric_name = next(iter(metric_map))
            CacheTSMetrics.metric_timeseries_cache[metric_name] = metric_map[
                metric_name
            ]

    def update_metric_timeseries_cache(self, latest_metric_timeseries_data, start_time):
        start_time = start_time.replace(minute=0, second=0, microsecond=0)
        if CacheTSMetrics.metric_timeseries_cache is None:
            return latest_metric_timeseries_data

        # Update data points
        for metric_timeseries in latest_metric_timeseries_data:
            cache_metric_data = CacheTSMetrics.metric_timeseries_cache[
                metric_timeseries
            ]
            recent_metric_data = latest_metric_timeseries_data[metric_timeseries]

            self.remove_old_data_points(
                cache_metric_data, metric_timeseries, start_time
            )
            self.update_cache_data_points(
                cache_metric_data, recent_metric_data, metric_timeseries
            )

        return CacheTSMetrics.metric_timeseries_cache

    @staticmethod
    def remove_old_data_points(cache_metric_data, metric_timeseries, start_time):
        for identifier in cache_metric_data:
            data_points_to_remove = 0
            cache_data_points = cache_metric_data[identifier]
            for data_map_index in range(len(cache_data_points)):
                data_map = cache_data_points[data_map_index]
                if next(iter(data_map)) < start_time:
                    data_points_to_remove += 1
                    _logger.debug(
                        f"Removing old {metric_timeseries} datapoints {data_map} for Org {identifier}"
                    )
                else:
                    break
            del cache_data_points[:data_points_to_remove]

    @staticmethod
    def update_cache_data_points(
        cache_metric_data, recent_metric_data, metric_timeseries
    ):
        for identifier in recent_metric_data:
            if cache_metric_data.get(identifier) is None:
                cache_metric_data[identifier] = recent_metric_data[identifier]
                _logger.info(
                    f"Adding {metric_timeseries} datapoints {recent_metric_data[identifier]} to cache "
                    f"for new Org {identifier}"
                )
            elif len(cache_metric_data.get(identifier)) == 0:
                cache_metric_data[identifier].extend(recent_metric_data[identifier])
                _logger.info(
                    f"No datapoints found in cache. Adding {metric_timeseries} datapoints {cache_metric_data[identifier]} "
                    f"for existing Org {identifier}"
                )
            else:
                recent_data_points = recent_metric_data[identifier]
                cache_data_points = cache_metric_data[identifier]

                first_recent_datapoint_index = 0

                # Get the last data point from the cache
                cache_data_map = cache_data_points[-1]
                cache_date_val = next(iter(cache_data_map))

                # Get the first data point from the recent query data
                recent_data_map = recent_data_points[first_recent_datapoint_index]
                recent_date_val = next(iter(recent_data_map))

                # if last datapoint time from cache is same as first datapoint update the cache datapoint
                if cache_date_val == recent_date_val:
                    if metric_timeseries == "active_endpoints":
                        if (
                            cache_data_map[cache_date_val]
                            < recent_data_map[recent_date_val]
                        ):
                            _logger.debug(
                                f"Found new max {metric_timeseries} datapoint {recent_data_map} replacing "
                                f"{cache_data_map} for Org {identifier}"
                            )
                            cache_data_map[cache_date_val] = recent_data_map[
                                recent_date_val
                            ]
                    else:
                        _logger.debug(
                            f"Adding up {metric_timeseries} datapoints {cache_data_map} with {recent_data_map} "
                            f"for Org {identifier}"
                        )
                        cache_data_map[cache_date_val] += recent_data_map[
                            recent_date_val
                        ]

                    first_recent_datapoint_index += 1

                # if last datapoint time is less than first datapoint add all new recent data points
                if 0 < len(recent_data_points[first_recent_datapoint_index:]):
                    cache_data_points.extend(
                        recent_data_points[first_recent_datapoint_index:]
                    )
                    _logger.debug(
                        f"Added new {metric_timeseries} datapoints {recent_data_points} for existing Org {identifier}"
                    )
