import datetime
import hashlib
import logging

from org_mapper.metrics_v2 import TimeStreamMetrics, CloudWatchMetrics, MongoMetrics
from org_mapper.transformer import Transformer

_logger = logging.getLogger(__name__)


class LCTransformer(Transformer):
    def __init__(self, period_back_metric_points, metric_backend, config):
        super().__init__(period_back_metric_points, metric_backend, config)
        # CloudWatch Metrics Specific to LC Backend
        self.metrics = [
            "ingress_events_cbth",
            "ingress_events_cbd",
            "ingress_events_cbd_long_term",
            "ingress_events_xdr",
        ]
        self.effective_traffic_config_key = config.get(
            "lc_effective_endpoints", "org_mapper_config_key"
        )
        self.manual_override_key = config.get(
            "lc_effective_endpoints", "manual_override_key"
        )
        self.manual_override_backup_key = config.get(
            "lc_effective_endpoints", "manual_override_backup_key"
        )
        self.manual_override_priority_key = config.get(
            "lc_effective_endpoints", "manual_override_priority_key"
        )
        self.manual_override_priority_backup_key = config.get(
            "lc_effective_endpoints", "manual_override_priority_backup_key"
        )
        self.bandwidth_units_per_partition = config.getint(
            "lc_endpoint_calculator", "effective_endpoints_per_partition"
        )
        self.large_org_longterm_min_effective_endpoints = config.getint(
            "lc_effective_endpoints", "large_org_longterm_min_effective_endpoints"
        )

        self.doc_builder_override_key = config.get(
            "lc_effective_endpoints", "doc_builder_override_key"
        )
        self.last_hour_higher_factor = config.getfloat(
            "lc_effective_endpoints", "last_hour_higher_factor"
        )
        self.save_effective_endpoint_cache = config.getboolean(
            "lc_effective_endpoints", "save_effective_endpoint_cache"
        )
        self.limit_bandwidth_double = config.getboolean(
            "lc_endpoint_calculator", "limit_effective_endpoint_double"
        )
        self.limit_bandwidth_reduction = config.getboolean(
            "lc_endpoint_calculator", "limit_effective_endpoint_reduction"
        )
        self.holiday_check = config.getboolean(
            "lc_endpoint_calculator",
            "limit_effective_endpoint_reduction_during_holiday_period",
        )
        self.bandwidth_reduction_factor = config.getfloat(
            "lc_endpoint_calculator", "effective_endpoint_reduction_factor"
        )
        self.minimum_update_time_threshold_hours = config.getint(
            "lc_endpoint_calculator", "minimum_update_time_threshold_hours"
        )
        self.minimum_reduction_factor = config.getfloat(
            "lc_endpoint_calculator", "minimum_reduction_factor"
        )
        self.minimum_increase_factor = config.getfloat(
            "lc_endpoint_calculator", "minimum_increase_factor"
        )

        self.mevents_per_partition = config.getint(
            "lc_endpoint_calculator", "mevents_per_partition"
        )
        self.max_effective_endpoints = config.getint(
            "lc_endpoint_calculator", "max_effective_endpoints"
        )
        self.cbd_to_cbth_multiplier = config.getint(
            "lc_endpoint_calculator", "cbd_to_cbth_multiplier"
        )
        if "CloudWatch" == metric_backend:
            self.metric_reader = CloudWatchMetrics(
                self.period_back_metric_points, self.config
            )
            self.metric_stats_list = [
                ("active_endpoints", "Maximum"),
                ("ingress_events_cbth", "Sum"),
                ("ingress_events_cbd", "Sum"),
                ("ingress_events_cbd_long_term", "Sum"),
                ("ingress_events_xdr", "Sum"),
            ]
        elif "TimeStream" == metric_backend:
            self.metric_reader = TimeStreamMetrics(
                self.period_back_metric_points, self.config, identifier_column="org_key"
            )
            # Do not alter the order of metrics
            self.metric_stats_list = [
                ("active_endpoints", "MAX"),
                ("ingress_events_cbd", "SUM"),
                ("ingress_events_cbth", "SUM"),
                ("ingress_events_cbd_long_term", "SUM"),
                ("ingress_events_xdr", "SUM"),
            ]
        elif "MongoDB" == metric_backend:
            self.metric_reader = MongoMetrics(
                self.period_back_metric_points, self.config
            )
            self.metric_stats_list = [
                ("active_endpoints", "$max"),
                ("ingress_events_cbd", "$sum"),
                ("ingress_events_cbth", "$sum"),
                ("ingress_events_cbd_long_term", "$sum"),
                ("ingress_events_xdr", "$sum"),
            ]

        self.use_quarter_hr = config.getboolean(
            "lc_effective_endpoints", "use_quarter_hr"
        )

    def _get_manual_overrides(self):
        manual_overrides = super()._get_manual_overrides()
        # Assume if no storage class is specified that it is the default one
        updated_overrides = {}
        for manual_override in manual_overrides:
            if len(manual_override.split(":")) == 1:
                updated_overrides[f"{manual_override}:default"] = manual_overrides[
                    manual_override
                ]
            else:
                updated_overrides[manual_override] = manual_overrides[manual_override]
        return updated_overrides

    def _get_manual_override_priority(self):
        manual_overrides = super()._get_manual_override_priority()
        # Assume if no storage class is specified that it is the default one
        updated_overrides = {}
        for manual_override in manual_overrides:
            if len(manual_override.split(":")) == 1:
                updated_overrides[f"{manual_override}:default"] = manual_overrides[
                    manual_override
                ]
            else:
                updated_overrides[manual_override] = manual_overrides[manual_override]
        return updated_overrides

    @staticmethod
    def _metrics_to_dict(series, active_endpoints):
        combined_metric_summary = {}
        for identifier in active_endpoints:
            combined_metric_summary[identifier] = {
                "active_endpoints": active_endpoints[identifier]
            }
            for entry in series:
                (metric_name, metric_summary) = entry
                combined_metric_summary[identifier][metric_name] = (
                    metric_summary[identifier] if identifier in metric_summary else 0
                )
        return combined_metric_summary

    def _aggregate_metric_results(self):
        top_sum_series = []
        recent_sum_series = []
        last_hour_sum_series = []

        if "CloudWatch" == self.metric_backend:
            series_results = self.metric_reader.get_metrics_series_data(
                self.metric_parallelism, self.metric_stats_list
            )
            # re-assigning the same value to sync implementation with timestream
            recent_series_results = series_results
            last_hr_series_results = series_results
        else:
            series_results = self.metric_reader.get_metrics_series_data(
                self.metric_stats_list
            )
            recent_series_results = self.metric_reader.get_recent_metrics_series_data(
                self.metric_stats_list[1:], self.last_x_points
            )

            if self.use_quarter_hr:
                start_time = self.metric_reader.end_time - datetime.timedelta(
                    minutes=15
                )
                last_hr_series_results = (
                    self.metric_reader.get_recent_metrics_series_data(
                        self.metric_stats_list[1:], 1, start_time, None, 900
                    )
                )
            else:
                last_hr_series_results = (
                    self.metric_reader.get_recent_metrics_series_data(
                        self.metric_stats_list[1:], 1
                    )
                )

        active_endpoints = self.metric_reader.get_max_series(
            series_results.get("active_endpoints")
        )

        for metric in self.metrics:
            top_sum_series.append(
                (
                    metric,
                    self.metric_reader.get_top_x_points(
                        series_results.get(metric), self.top_x_points
                    ),
                )
            )
            recent_sum_series.append(
                (
                    metric,
                    self.metric_reader.get_last_x_points(
                        recent_series_results.get(metric), self.last_x_points
                    ),
                )
            )
            last_hour_sum_series.append(
                (
                    metric,
                    self.metric_reader.get_last_x_points(
                        last_hr_series_results.get(metric), 4
                    )
                    if self.use_quarter_hr
                    else self.metric_reader.get_last_x_points(
                        last_hr_series_results.get(metric), 1
                    ),
                )
            )

        top_combined_metric_summary = self._metrics_to_dict(
            top_sum_series, active_endpoints
        )
        recent_combined_metric_summary = self._metrics_to_dict(
            recent_sum_series, active_endpoints
        )
        last_hour_combined_metric_summary = self._metrics_to_dict(
            last_hour_sum_series, active_endpoints
        )

        return {
            "active_endpoints": active_endpoints,
            "top_combined_sum_series": top_combined_metric_summary,
            "recent_combined_sum_series": recent_combined_metric_summary,
            "last_hour_combined_sum_series": last_hour_combined_metric_summary,
        }

    def convert_bandwidth_shard_map(self):
        aggr_metric_results = self._aggregate_metric_results()
        active_endpoints = aggr_metric_results.get("active_endpoints")

        self._get_config()
        old_effective_endpoints = self.get_total_effective_traffic()
        remove_list = self._get_identifiers_to_delete(active_endpoints)
        self._remove_ids(remove_list)

        top_combined_sum_series = aggr_metric_results.get("top_combined_sum_series")
        recent_combined_sum_series = aggr_metric_results.get(
            "recent_combined_sum_series"
        )
        last_hour_combined_sum_series = aggr_metric_results.get(
            "last_hour_combined_sum_series"
        )

        for org in active_endpoints:
            default_org = f"{org}:default"
            longterm_org = f"{org}:longTerm"
            default_old_value = None
            longterm_old_value = None

            if default_org in self.effective_traffic:
                default_old_value = self._get_id_effective_traffic_value(default_org)

            if longterm_org in self.effective_traffic:
                longterm_old_value = self._get_id_effective_traffic_value(longterm_org)

            (
                default_top_ee_result,
                longterm_top_ee_result,
            ) = self.calculate_effective_endpoints(
                top_combined_sum_series[org]["ingress_events_cbth"],
                top_combined_sum_series[org]["ingress_events_cbd"],
                top_combined_sum_series[org]["ingress_events_cbd_long_term"],
                top_combined_sum_series[org]["ingress_events_xdr"],
                top_combined_sum_series[org]["active_endpoints"],
                org,
                period_back_metric_points=self.top_x_points,
            )

            (
                default_recent_ee_result,
                longterm_recent_ee_result,
            ) = self.calculate_effective_endpoints(
                recent_combined_sum_series[org]["ingress_events_cbth"],
                recent_combined_sum_series[org]["ingress_events_cbd"],
                recent_combined_sum_series[org]["ingress_events_cbd_long_term"],
                recent_combined_sum_series[org]["ingress_events_xdr"],
                recent_combined_sum_series[org]["active_endpoints"],
                org,
                period_back_metric_points=self.last_x_points,
            )

            (
                default_last_hour_ee_result,
                longterm_last_hour_ee_result,
            ) = self.calculate_effective_endpoints(
                last_hour_combined_sum_series[org]["ingress_events_cbth"],
                last_hour_combined_sum_series[org]["ingress_events_cbd"],
                last_hour_combined_sum_series[org]["ingress_events_cbd_long_term"],
                last_hour_combined_sum_series[org]["ingress_events_xdr"],
                last_hour_combined_sum_series[org]["active_endpoints"],
                org,
                period_back_metric_points=1,
            )

            default_ee_result = max(default_top_ee_result, default_recent_ee_result)
            longterm_ee_result = max(longterm_top_ee_result, longterm_recent_ee_result)

            default_ee_result = self._use_last_hour(
                default_ee_result, default_last_hour_ee_result, default_old_value
            )
            longterm_ee_result = self._use_last_hour(
                longterm_ee_result, longterm_last_hour_ee_result, longterm_old_value
            )

            if (
                (default_recent_ee_result == default_ee_result)
                or (default_last_hour_ee_result == default_ee_result)
            ) and default_ee_result != default_top_ee_result:
                _logger.info(
                    f"Org: {org}, 7 day Look back: {default_top_ee_result}, "
                    f"4 hour Look back: {default_recent_ee_result}, "
                    f"Last hour: {default_last_hour_ee_result}, Previous Value: {default_old_value}, "
                    f"Using: {default_ee_result}\n"
                    f"longterm: 7 day Look back: {longterm_top_ee_result}, "
                    f"4 hour Look back: {longterm_recent_ee_result}, "
                    f"Last hour: {longterm_last_hour_ee_result}, Previous Value: {longterm_old_value}, "
                    f"Using: {longterm_ee_result}"
                )

            default_ee_result = self.limit_bandwidth_movement(
                default_ee_result, default_old_value, org
            )
            longterm_ee_result = self.limit_bandwidth_movement(
                longterm_ee_result, longterm_old_value, org
            )

            last_updated_time = self._get_last_update_time(default_org)
            endpoint_value = self._get_id_effective_traffic_value(default_org)

            self._update_effective_endpoints(
                default_ee_result, last_updated_time, default_org, endpoint_value
            )

            last_updated_time = self._get_last_update_time(longterm_org)
            endpoint_value = self._get_id_effective_traffic_value(longterm_org)

            #  Special case to set long term to zero on large orgs when
            #  below a configurable threshold of effective.
            #  This is so we don't have a separate shard for long term in those orgs.
            if (
                default_ee_result > self.bandwidth_units_per_partition
                and longterm_ee_result < self.large_org_longterm_min_effective_endpoints
            ):
                longterm_ee_result = 0

            self._update_effective_endpoints(
                longterm_ee_result, last_updated_time, longterm_org, endpoint_value
            )

            # TODO: this probably can be removed it was here for when we transitioned from org to org:default to remove
            #  all the entries that had just org.
            if org in self.effective_traffic.keys():
                _logger.info(f"Removing {org} from the org mapper")
                del self.effective_traffic[org]

        if self.save_effective_endpoint_cache:
            self._save_config()

        _logger.info(
            f"Effective Endpoints Total, old: {old_effective_endpoints}, new: {self.get_total_effective_traffic()}"
        )

    def _get_identifiers_to_delete(self, active_endpoints):
        remove_list = []
        for org_class in self.effective_traffic:
            if len(org_class.split(":")) > 1:
                org, storage_class = org_class.split(":")
            else:
                org = org_class
            if org not in active_endpoints:
                remove_list.append(org_class)

        return remove_list

    @staticmethod
    def _generate_kinesis_hash(org, partitionid):
        orgid, storage_class = org.split(":")
        # Hash key is the first 11 bytes of the MD5 of the org id,
        # with the first 5 bytes of the storage_class:partitionid appended.
        m = hashlib.new("md5", usedforsecurity=False)
        m.update(bytes(orgid, "utf-8"))
        hash_org = m.digest()[:11]
        # re-initialize hashlib Python update appends the new data
        m = hashlib.new("md5", usedforsecurity=False)
        m.update(bytes(f"{storage_class}:{partitionid}", "utf-8"))
        hash_storage_class = m.digest()[:5]
        hash_key = int.from_bytes(hash_org + hash_storage_class, "big")
        return hash_key

    def _update_effective_endpoints(
        self, default_ee_result, last_updated_time, org, endpoint_value
    ):
        if self.update_bandwidth_rules(
            default_ee_result, last_updated_time, org, endpoint_value
        ):
            if default_ee_result != endpoint_value:
                _logger.info(
                    f"Updating {org} from {endpoint_value} to {default_ee_result}"
                )
            self._update_id_effective_traffic(org, default_ee_result)
        else:
            if self._get_last_update_time(org):
                last_updated_string = datetime.datetime.fromtimestamp(
                    last_updated_time
                ).strftime("%c")
            else:
                last_updated_string = "Never"

            _logger.debug(
                f"Rule prevented {org} from updating to {default_ee_result}.  "
                f"Old Value {endpoint_value} "
                f"Last Updated {last_updated_string}"
            )

    def _class_calculate_effective_endpoints(
        self, adjusted_event_rate, active_endpoints, org="UNKNOWN"
    ):
        effective_endpoints = (
            adjusted_event_rate
            * self.bandwidth_units_per_partition
            / self.mevents_per_partition
            / 1000000.0
        )

        return effective_endpoints

    def calculate_effective_endpoints(
        self,
        threathunter_rate,
        defense_rate,
        defense_longterm_rate,
        xdr_rate,
        active_endpoints,
        org="UNKNOWN",
        period_back_metric_points=24,
    ):
        _logger.debug(
            f"TH Rate: {threathunter_rate}, CBD Rate: {defense_rate}, XDR Rate: {xdr_rate}, "
            f"Active_Endpoints {active_endpoints}"
        )
        default_adjusted_rate = (
            threathunter_rate + xdr_rate + defense_rate * self.cbd_to_cbth_multiplier
        )
        # Above rate assumes 24 hour window, pro-rate for smaller windows.
        default_adjusted_rate = (default_adjusted_rate / period_back_metric_points) * 24
        default_effective_endpoints = self._class_calculate_effective_endpoints(
            default_adjusted_rate, active_endpoints, f"{org}:default"
        )
        longterm_adjusted_rate = defense_longterm_rate * self.cbd_to_cbth_multiplier
        longterm_adjusted_rate = (
            longterm_adjusted_rate / period_back_metric_points
        ) * 24
        longterm_effective_endpoints = self._class_calculate_effective_endpoints(
            longterm_adjusted_rate, active_endpoints, f"{org}:longTerm"
        )
        return default_effective_endpoints, longterm_effective_endpoints
