import logging
import time
from datetime import datetime, timedelta
from itertools import zip_longest
from multiprocessing.pool import ThreadPool
from collections import defaultdict

import pymongo
from pymongo import MongoClient
from pscr_common.aws import get_boto3_client

from org_mapper import timestream_util
from org_mapper.cache_ts_metrics import CacheTSMetrics
from org_mapper.timestream_util import execute_query, build_query
from org_mapper.utils import log_and_sleep, str_to_bool

_logger = logging.getLogger(__name__)


class Metrics:
    def __init__(self, look_back_metric_points, config):
        self.namespace = config.get("env", "id") + "_custom"
        self.look_back_metric_points = look_back_metric_points
        self.cache_timeseries = {}

    def get_last_x_points(self, metric_series, last_x_points):
        raise NotImplemented

    def get_top_x_points(self, metric_series, top_x_points):
        raise NotImplemented


class CloudWatchMetrics(Metrics):
    cw = get_boto3_client("cloudwatch")

    def __init__(self, look_back_metric_points, config):
        self.metric_period = config.getint("cloudwatch_metrics", "metric_period")
        now = time.time()
        self.start_time = datetime.utcfromtimestamp(
            now - (self.metric_period * look_back_metric_points)
        )
        self.end_time = datetime.utcfromtimestamp(now)
        super().__init__(look_back_metric_points, config)

    @staticmethod
    def get_top_x_points(metric_series, top_x_points, pro_rate=True):
        # Return sum of the top x metric points over the look_back_metric_points range.
        # Useful if you want to scale to highest data points over a long period of time.
        # So that you don't churn up and down a lot.
        highest_x_series = {}
        for org in metric_series:
            top_x_values = sorted(metric_series[org]["Values"])[-top_x_points:]
            highest_x_org = sum(top_x_values)
            number_points = len(top_x_values)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < top_x_points:
                    highest_x_org *= 1.0 * top_x_points / number_points
            highest_x_series[org] = highest_x_org
            _logger.debug(f"org: {org}, sum: {highest_x_org}, values; {top_x_values}")
        return highest_x_series

    @staticmethod
    def get_last_x_points(metric_series, last_x_points, pro_rate=True):
        sum_series = {}
        for org in metric_series:
            #  Cloudwatch returns values ordered from most recent to oldest, so we want the first items on the list
            last_x_values = metric_series[org]["Values"][:last_x_points]
            sum_org = sum(last_x_values)
            number_points = len(last_x_values)
            if pro_rate:
                if 0 < number_points < last_x_points:
                    sum_org *= 1.0 * last_x_points / number_points
            sum_series[org] = sum_org
        return sum_series

    def get_metrics_series_data(self, metric_parallelism, metric_stats_list):
        """
        This function collects metrics from cloud watch for given metric names (metric list)
        :param metric_parallelism: No of Thread Pool Processes
        :param metric_stats_list: list of maps of cloudwatch metric names and stat
        :return: cw_metrics_series_results
        """
        start_time = datetime.now()
        p = ThreadPool(processes=metric_parallelism)
        metric_series_results = dict(
            p.starmap(self.get_metric_series_with_context, metric_stats_list)
        )
        p.close()
        ms = datetime.now() - start_time
        _logger.info(f"Collected all metric in {ms}")
        return metric_series_results

    @staticmethod
    def get_max_series(series, remove_zeros=True):
        max_series = {}
        for org in series:
            number_values = len(series[org]["Values"])
            if number_values == 0 and not remove_zeros:
                max_series[org] = 0
            elif number_values > 0:
                if org in max_series:
                    if series[org]["Values"] > max_series[org]:
                        max_series[org] = max(series[org]["Values"])
                else:
                    max_series[org] = max(series[org]["Values"])
        return max_series

    @staticmethod
    def _parse_metric_results(series, response):
        for results in response["MetricDataResults"]:
            if results["StatusCode"] == "Complete":
                #  Split off the `a_` we prepended to org ID in the Id field above.
                #  AWS requires first character to be a letter
                org = results["Id"].split("_")[1]
                series[org] = {
                    "Timestamps": results["Timestamps"],
                    "Values": results["Values"],
                }

    @staticmethod
    def _grouper(inputs, n, fill_value=None):
        # return first n as a tuple, then next n, etc.  Pad final set with None
        iters = [iter(inputs)] * n
        return zip_longest(*iters, fillvalue=fill_value)

    def _get_dimensions(self, metric_name):
        _logger.info(f"Getting Dimensions for {metric_name}")
        dimensions = []
        response = self._get_aws_dimensions(
            {"Namespace": self.namespace, "MetricName": metric_name}
        )
        for metric in response["Metrics"]:
            dimensions += metric["Dimensions"]
        while "NextToken" in response:
            response = self._get_aws_dimensions(
                {
                    "Namespace": self.namespace,
                    "MetricName": metric_name,
                    "NextToken": response.get("NextToken"),
                }
            )
            for metric in response["Metrics"]:
                dimensions += metric["Dimensions"]
        return dimensions

    def _get_aws_dimensions(self, kwargs):
        retries = 0
        while True:
            try:
                response = self.cw.list_metrics(**kwargs)
            except self.cw.exceptions.ClientError as e:
                log_and_sleep(
                    f"Cloudwatch Exception retrying: {e}", retries, max_sleep=60
                )
            else:
                break
            retries += 1

        return response

    def get_metric_series_with_context(self, metric_name, aws_stat="Sum", cache=False):
        return metric_name, self.get_metric_series(metric_name, aws_stat, cache)

    def get_metric_series(self, metric_name, aws_stat="Sum", cache=False):
        _logger.info(f"Getting Metric series for {metric_name}")
        metric_id = metric_name + "-" + aws_stat
        if metric_id in self.cache_timeseries:
            return self.cache_timeseries[metric_id]

        dimensions = self._get_dimensions(metric_name)
        if len(dimensions) == 0:
            return {}
        series = self._metric_get_helper(dimensions, metric_name, aws_stat)

        if cache:
            self.cache_timeseries[metric_id] = series

        return series

    def _metric_get_helper(self, dimensions, metric_name, aws_stat):
        series = {}
        group_dimensions = self._grouper(dimensions, 100)
        for hundred_dimensions in group_dimensions:
            request_payload = []
            for dimension in hundred_dimensions:
                if dimension:
                    #  Defensive code, in dev we had some bad org names with dashes,  That shouldn't be a valid orgkey
                    #  And is invalid to insert into the Id field here
                    if "-" not in dimension["Value"]:
                        metric_request = {
                            "Id": f"a_{dimension['Value']}",
                            "MetricStat": {
                                "Metric": {
                                    "Namespace": self.namespace,
                                    "MetricName": metric_name,
                                    "Dimensions": [dimension],
                                },
                                "Period": self.metric_period,
                                "Stat": aws_stat,
                            },
                        }
                        request_payload.append(metric_request)
            response = self._get_aws_metric_series(
                {
                    "MetricDataQueries": request_payload,
                    "StartTime": self.start_time,
                    "EndTime": self.end_time,
                }
            )
            self._parse_metric_results(series, response)

            while "NextToken" in response:
                response = self._get_aws_metric_series(
                    {
                        "MetricDataQueries": request_payload,
                        "StartTime": self.start_time,
                        "EndTime": self.end_time,
                        "NextToken": response.get("NextToken"),
                    }
                )
                self._parse_metric_results(series, response)

        return series

    def _get_aws_metric_series(self, kwargs):
        retries = 0
        while True:
            try:
                response = self.cw.get_metric_data(**kwargs)
                _logger.debug("Success getting metric series")
            except self.cw.exceptions.ClientError as ex:
                log_and_sleep(
                    f"Cloudwatch Limit Exceeded Exception for get metric data retrying: {ex}",
                    retries,
                    max_sleep=60,
                )
            else:
                break
            retries += 1

        return response

    @staticmethod
    def get_avg_series(series):
        avg_series = {}
        for org in series:
            number_values = len(series[org]["Values"])
            if number_values == 0:
                avg_series[org] = 0
            else:
                avg_series[org] = sum(series[org]["Values"]) / number_values
        return avg_series


class TimeStreamMetrics(Metrics):
    timestream = get_boto3_client("timestream-query")
    prev_run_time = None

    def __init__(
        self,
        look_back_metric_points,
        config,
        identifier_column="org_key",
        metric_period=None,
    ):
        self.database = config.get("timestream", "db_name")
        self.table = config.get("timestream", "table_name")
        self.env = config.get("env", "id")
        if metric_period:
            self.metric_period = metric_period
        else:
            self.metric_period = config.getint("timestream", "metric_period")
        now = time.time()
        self.start_time = datetime.utcfromtimestamp(
            now - (self.metric_period * look_back_metric_points)
        )
        self.end_time = datetime.utcfromtimestamp(now)
        self.identifier_column = identifier_column
        self.cache = str_to_bool(config.get("timestream", "cache"))
        self.curr_day_start_time = datetime.utcnow().replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        self.cache_ts_metrics = CacheTSMetrics(
            look_back_metric_points, config, self.identifier_column
        )
        self.paginator = self.timestream.get_paginator("query")
        super().__init__(look_back_metric_points, config)
        TimeStreamMetrics.prev_run_time = self.end_time

    def get_metrics_series_data(self, metric_list):
        metric_timeseries_data = []
        metric_timeseries_data_map = {}

        for metrics in metric_list:
            metric_timeseries_data.append({metrics[0]: {}})

        start_time = self.start_time

        if self.cache:
            if CacheTSMetrics.metric_timeseries_cache is None:
                self.cache_ts_metrics.load_metric_timeseries_to_cache(
                    self.paginator, start_time, self.end_time, metric_list
                )
                return CacheTSMetrics.metric_timeseries_cache
            else:
                if timestream_util.prev_run_time is None:
                    raise Exception("Prev Run Time is None while cache exist")

                start_time = timestream_util.prev_run_time

        query = build_query(
            self.identifier_column,
            self.metric_period,
            self.env,
            self.database,
            self.table,
            start_time,
            self.end_time,
            metric_list,
        )

        _logger.debug(f"Timestream Query: {query}")

        execute_query(self.paginator, query, metric_timeseries_data)

        for metric_map in metric_timeseries_data:
            metric_name = next(iter(metric_map))
            metric_timeseries_data_map[metric_name] = metric_map[metric_name]

        return self.cache_ts_metrics.update_metric_timeseries_cache(
            metric_timeseries_data_map, self.start_time
        )

    def get_recent_metrics_series_data(
        self,
        metric_list,
        last_x_points,
        start_time=None,
        end_time=None,
        metric_period=None,
    ):
        metric_timeseries_data = []
        metric_timeseries_data_map = {}

        for metrics in metric_list:
            metric_timeseries_data.append({metrics[0]: {}})

        if not end_time:
            end_time = self.end_time

        if not start_time:
            start_time = self.end_time - timedelta(hours=last_x_points)

        if not metric_period:
            metric_period = self.metric_period

        query = build_query(
            self.identifier_column,
            metric_period,
            self.env,
            self.database,
            self.table,
            start_time,
            end_time,
            metric_list,
        )

        _logger.debug(f"Timestream Query: {query}")

        execute_query(self.paginator, query, metric_timeseries_data)

        for metric_map in metric_timeseries_data:
            metric_name = next(iter(metric_map))
            metric_timeseries_data_map[metric_name] = metric_map[metric_name]

        return metric_timeseries_data_map

    @staticmethod
    def get_max_series(series):
        max_series = {}
        for identifier in series:
            max_data_points = [next(iter(d.values())) for d in series[identifier]]
            if len(max_data_points) > 0:
                max_series[identifier] = max(max_data_points)

        return max_series

    @staticmethod
    def get_top_x_points(series, top_x_points, pro_rate=True):
        top_sum_series = dict()
        for identifier in series:
            top_x_data_points = sorted(
                [next(iter(d.values())) for d in series[identifier]]
            )[-top_x_points:]
            highest_x_org = sum(top_x_data_points)
            number_points = len(top_x_data_points)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < top_x_points:
                    highest_x_org *= 1.0 * top_x_points / number_points

            top_sum_series[identifier] = highest_x_org

        return top_sum_series

    def get_last_x_points(self, series, last_x_points, pro_rate=True):
        last_sum_series = dict()
        for identifier in series:
            last_x_data_points = [next(iter(d.values())) for d in series[identifier]]
            highest_x_org = sum(last_x_data_points)
            number_points = len(last_x_data_points)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < last_x_points:
                    highest_x_org *= 1.0 * last_x_points / number_points

            last_sum_series[identifier] = highest_x_org

        return last_sum_series


class MongoMetrics(Metrics):
    def __init__(
        self,
        look_back_metric_points,
        config,
        identifier_column="org_key",
        timestamp_column="timestamp",
        metric_period=None,
    ):
        self.env = config.get("env", "id")
        (
            self.mongo_collection,
            self.mongo_db,
            self.mongo_admin,
        ) = self._init_mongo_clients(config)

        if metric_period:
            self.metric_period = metric_period
        else:
            self.metric_period = config.getint("mongo", "metric_period")
        now = time.time()
        self.start_time = datetime.utcfromtimestamp(
            now - (self.metric_period * look_back_metric_points)
        )
        self.end_time = datetime.utcfromtimestamp(now)
        self.identifier_column = identifier_column
        self.timestamp_column = timestamp_column
        self.curr_day_start_time = datetime.utcnow().replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        super().__init__(look_back_metric_points, config)

    def _init_mongo_clients(self, config):
        host = config.get("mongo", "host")
        port = config.get("mongo", "port")
        user = config.get("mongo", "user")
        password = config.get("mongo", "password")
        db_name = config.get("mongo", "db_name")
        self.collection_name = config.get("mongo", "collection_name")
        if user == "none":
            uri = f"mongodb://{host}:{port}"
        else:
            # uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}?directConnection=true"
            uri = f"mongodb://{user}:{password}@{host}:{port}/{db_name}"
        client = MongoClient(uri)

        return client[db_name][self.collection_name], client[db_name], client.admin

    def init_db(self):
        """Create the timeseries collection (used for UTs, not real envs)"""
        self.mongo_db.create_collection(
            self.collection_name,
            timeseries={
                "timeField": self.timestamp_column,
                "metaField": self.identifier_column,
                "granularity": "hours",
            },
        )

    def _get_metrics_series_data(
        self,
        metric_list: list[tuple[str, str]],
        start_time: datetime,
        end_time: datetime,
        metric_period: int,
    ) -> dict:
        """Aggregate the given metrics data

        Args:
            metric_list (list[tuple[str, str]]): tuples of the metric names and the function used to aggregate them
            start_time (datetime): start of period
            end_time (datetime): end of period
            metric_period (int): bin size in seconds

        Returns:
            dict: Aggregated metrics
        """
        metric_aggregations = {m[0]: {m[1]: f"${m[0]}"} for m in metric_list}

        agg_pipeline = [
            # Stage 1 - match the time period
            {"$match": {self.timestamp_column: {"$gte": start_time, "$lt": end_time}}},
            # Stage 2 - group by identifier column (org_key/subnet_id) and by the binned period, and do the metric aggregations
            {
                "$group": {
                    "_id": {
                        self.identifier_column: f"${self.identifier_column}",
                        # bin time
                        f"{self.timestamp_column}_bin": {
                            "$dateTrunc": {
                                "date": f"${self.timestamp_column}",
                                "unit": "second",
                                "binSize": metric_period,
                            }
                        },
                    },
                    # add in the selected metric aggregations
                    **metric_aggregations,
                }
            },
            # Stage 3 - sort by identifier column and binned time
            {"$sort": {"_id": 1}},
        ]

        _logger.debug("Mongo aggregation pipeline: %s", agg_pipeline)
        metric_timeseries_data_map = {}

        for metric in metric_list:
            metric_timeseries_data_map[metric[0]] = defaultdict(list)

        for r in self.mongo_collection.aggregate(agg_pipeline):
            org_key = r["_id"][self.identifier_column]
            timestamp = r["_id"][f"{self.timestamp_column}_bin"]
            for metric in metric_list:
                metric_timeseries_data_map[metric[0]][org_key].append(
                    {timestamp: r[metric[0]] or 0}
                )

        return metric_timeseries_data_map

    def get_metrics_series_data(self, metric_list):
        return self._get_metrics_series_data(
            metric_list, self.start_time, self.end_time, self.metric_period
        )

    def get_recent_metrics_series_data(
        self,
        metric_list,
        last_x_points,
        start_time=None,
        end_time=None,
        metric_period=None,
    ):
        end_time = end_time or self.end_time

        if not start_time:
            start_time = self.end_time - timedelta(hours=last_x_points)

        metric_period = metric_period or self.metric_period

        return self._get_metrics_series_data(
            metric_list, start_time, end_time, metric_period
        )

    @staticmethod
    def get_max_series(series):
        max_series = {}
        for identifier in series:
            max_data_points = [next(iter(d.values())) for d in series[identifier]]
            if len(max_data_points) > 0:
                max_series[identifier] = max(max_data_points)

        return max_series

    @staticmethod
    def get_top_x_points(series, top_x_points, pro_rate=True):
        top_sum_series = dict()
        for identifier in series:
            top_x_data_points = sorted(
                [next(iter(d.values())) for d in series[identifier]]
            )[-top_x_points:]
            highest_x_org = sum(top_x_data_points)
            number_points = len(top_x_data_points)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < top_x_points:
                    highest_x_org *= 1.0 * top_x_points / number_points

            top_sum_series[identifier] = highest_x_org

        return top_sum_series

    def get_last_x_points(self, series, last_x_points, pro_rate=True):
        last_sum_series = dict()
        for identifier in series:
            last_x_data_points = [next(iter(d.values())) for d in series[identifier]]
            highest_x_org = sum(last_x_data_points)
            number_points = len(last_x_data_points)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < last_x_points:
                    highest_x_org *= 1.0 * last_x_points / number_points

            last_sum_series[identifier] = highest_x_org

        return last_sum_series
