import datetime
import hashlib
import logging
from org_mapper.metrics_v2 import TimeStreamMetrics, CloudWatchMetrics, MongoMetrics
from org_mapper.transformer import Transformer

_logger = logging.getLogger(__name__)


class NTATransformer(Transformer):
    def __init__(self, period_back_metric_points, metric_backend, config):
        super().__init__(period_back_metric_points, metric_backend, config)
        self.metric = "nta_netconn_event_count"
        self.cloudwatch_metric_stat = ("nta_netconn_event_count", "Sum")
        self.ts_metric_stat = ("nta_netconn_event_count", "SUM")
        self.mongo_metric_stat = ("nta_netconn_event_count", "$sum")

        self.effective_traffic_config_key = config.get(
            "nta_effective_bandwidth", "org_mapper_config_key"
        )
        self.manual_override_key = config.get(
            "nta_effective_bandwidth", "manual_override_key"
        )
        self.manual_override_backup_key = config.get(
            "nta_effective_bandwidth", "manual_override_backup_key"
        )
        self.manual_override_priority_key = config.get(
            "nta_effective_bandwidth", "manual_override_priority_key"
        )
        self.manual_override_priority_backup_key = config.get(
            "nta_effective_bandwidth", "manual_override_priority_backup_key"
        )
        self.bandwidth_units_per_partition = config.getint(
            "nta_effective_bandwidth", "bandwidth_units_per_partition"
        )
        self.doc_builder_override_key = config.get(
            "nta_effective_bandwidth", "doc_builder_override_key"
        )
        self.last_hour_higher_factor = config.getfloat(
            "nta_effective_bandwidth", "last_hour_higher_factor"
        )
        self.save_config = config.getboolean("nta_effective_bandwidth", "save_config")
        self.limit_bandwidth_double = config.getboolean(
            "nta_effective_bandwidth", "limit_bandwidth_double"
        )
        self.limit_bandwidth_reduction = config.getboolean(
            "nta_effective_bandwidth", "limit_bandwidth_reduction"
        )
        self.holiday_check = config.getboolean(
            "nta_effective_bandwidth", "limit_bandwidth_reduction_during_holiday_period"
        )
        self.bandwidth_reduction_factor = config.getfloat(
            "nta_effective_bandwidth", "bandwidth_reduction_factor"
        )
        self.minimum_update_time_threshold_hours = config.getint(
            "nta_effective_bandwidth", "minimum_update_time_threshold_hours"
        )
        self.minimum_reduction_factor = config.getfloat(
            "nta_effective_bandwidth", "minimum_reduction_factor"
        )
        self.minimum_increase_factor = config.getfloat(
            "nta_effective_bandwidth", "minimum_increase_factor"
        )

        self.mevents_per_partition = config.getfloat(
            "nta_effective_bandwidth", "mevents_per_partition"
        )
        self.max_bandwidth_units = config.getint(
            "nta_effective_bandwidth", "max_bandwidth_units"
        )
        if "CloudWatch" == metric_backend:
            self.metric_reader = CloudWatchMetrics(
                self.period_back_metric_points, self.config
            )
            self.metric_stat = self.cloudwatch_metric_stat
        elif "TimeStream" == metric_backend:
            self.metric_reader = TimeStreamMetrics(
                self.period_back_metric_points,
                self.config,
                identifier_column="subnet_id",
            )
            self.metric_stat = self.ts_metric_stat
        elif "MongoDB" == metric_backend:
            self.metric_reader = TimeStreamMetrics(
                self.period_back_metric_points,
                self.config,
                identifier_column="subnet_id",
            )
            self.metric_stat = self.mongo_metric_stat

    def _metric_results(self):
        if "CloudWatch" in self.metric_backend:
            series_results = self.metric_reader.get_metrics_series_data(
                self.metric_parallelism, [self.metric_stat]
            )
            # re-assigning the same value to sync implementation with timestream
            recent_series_results = series_results
            last_hr_series_results = series_results
        else:
            series_results = self.metric_reader.get_metrics_series_data(
                [self.metric_stat]
            )
            recent_series_results = self.metric_reader.get_recent_metrics_series_data(
                [self.metric_stat], self.last_x_points
            )
            last_hr_series_results = self.metric_reader.get_recent_metrics_series_data(
                [self.metric_stat], 1
            )

        top_sum_series = self.metric_reader.get_top_x_points(
            series_results.get(self.metric), self.top_x_points
        )
        recent_sum_series = self.metric_reader.get_last_x_points(
            recent_series_results.get(self.metric), self.last_x_points
        )
        last_hour_sum_series = self.metric_reader.get_last_x_points(
            last_hr_series_results.get(self.metric), 1
        )

        return top_sum_series, recent_sum_series, last_hour_sum_series

    def convert_bandwidth_shard_map(self):
        self._get_config()
        old_bandwidth_units = self.get_total_effective_traffic()
        top_sum_series, recent_sum_series, last_hour_sum_series = self._metric_results()
        ids = top_sum_series.keys()
        remove_list = self._get_identifiers_to_delete(ids)
        self._remove_ids(remove_list)

        for identifier in ids:
            if identifier in self.effective_traffic:
                old_value = self._get_id_effective_traffic_value(identifier)
            else:
                old_value = None

            top_bandwidth_units = self.calculate_bandwidth_units(
                top_sum_series[identifier], self.top_x_points, identifier
            )
            if identifier in recent_sum_series:
                recent_bandwidth_units = self.calculate_bandwidth_units(
                    recent_sum_series[identifier], self.last_x_points, identifier
                )
            else:
                recent_bandwidth_units = 0
            if identifier in last_hour_sum_series:
                last_hour_bandwidth_units = self.calculate_bandwidth_units(
                    last_hour_sum_series[identifier], 1, identifier
                )
            else:
                last_hour_bandwidth_units = 0

            bandwidth_units = max(top_bandwidth_units, recent_bandwidth_units)
            bandwidth_units = self._use_last_hour(
                bandwidth_units, last_hour_bandwidth_units, old_value
            )

            if (
                (bandwidth_units == recent_bandwidth_units)
                or (bandwidth_units == last_hour_bandwidth_units)
            ) and bandwidth_units != top_bandwidth_units:
                _logger.info(
                    f"{identifier}, 7 day Look back: {top_bandwidth_units}, "
                    f"4 hour Look back: {recent_bandwidth_units}, "
                    f"Last hour: {last_hour_bandwidth_units}, Previous Value: {old_value}, "
                    f"Using: {bandwidth_units}\n"
                )

            bandwidth_units = self.limit_bandwidth_movement(
                bandwidth_units, old_value, identifier
            )
            last_updated_time = self._get_last_update_time(identifier)

            if self.update_bandwidth_rules(
                bandwidth_units, last_updated_time, identifier, old_value
            ):
                if bandwidth_units != old_value:
                    _logger.info(
                        f"Updating {identifier} from {old_value} to {bandwidth_units}"
                    )
                    self._update_id_effective_traffic(identifier, bandwidth_units)
            else:
                if last_updated_time:
                    last_updated_string = datetime.datetime.fromtimestamp(
                        last_updated_time
                    ).strftime("%c")
                else:
                    last_updated_string = "Never"
                _logger.debug(
                    f"Rule prevented {identifier} from updating to {bandwidth_units}.  "
                    f"Old Value {old_value} "
                    f"Last Updated {last_updated_string}"
                )
        if self.save_config:
            self._save_config()
        _logger.info(
            f"Bandwidth Units Total, old: {old_bandwidth_units}, new: {self.get_total_effective_traffic()}"
        )

        self._update_partition_keys()

    def calculate_bandwidth_units(
        self, metrics_sum, metric_points, identifier="UNKNOWN"
    ):
        # Divide by period metric points to get hourly rate
        adjusted_rate = metrics_sum / metric_points
        _logger.debug(f"Rate for {identifier}: {adjusted_rate}")
        bandwidth_units = (
            adjusted_rate
            * self.bandwidth_units_per_partition
            / self.mevents_per_partition
            / 1000000.0
        )
        if bandwidth_units > self.max_bandwidth_units:
            _logger.warning(
                f"Indicated bandwidth greater than max setting to {self.max_bandwidth_units}"
                f"for {identifier}"
            )
            bandwidth_units = self.max_bandwidth_units
        return round(bandwidth_units)

    @staticmethod
    def _generate_kinesis_hash(identifier, partitionid):
        # Hash key is the first 11 bytes of the subnet hash,
        # with the first 5 bytes of the hash of the partitionid appended.
        org, subnet_hash = identifier.split("|")
        subnet_id_bytes = bytes.fromhex(subnet_hash)
        hash_id = subnet_id_bytes[:11]
        m = hashlib.new("md5", usedforsecurity=False)
        m.update(bytes(f"{partitionid}", "utf-8"))
        partition_id_hash = m.digest()[:5]
        hash_key = int.from_bytes(hash_id + partition_id_hash, "big")
        return hash_key
