from datetime import datetime, timedelta, timezone
import time
import math
import logging
import os
import schedule
from pscr_common.log import setup_logging, clear_logging_handlers
from org_mapper.metrics_v2 import TimeStreamMetrics, MongoMetrics
from org_mapper.utils import Config, str_to_bool
from pscr_common.metrics import get_reporter, PrometheusRegistry
from pscr_common.metadata import MetadataProvider

logging.getLogger("boto").setLevel(logging.CRITICAL)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logging.getLogger("s3transfer").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.CRITICAL)
_logger = logging.getLogger(__name__)


class IdentifierStatsReporter:
    def __init__(self, config, metrics_obj=None):
        self.config = config
        self.mongodb_enabled = str_to_bool(self.config.get("mongo", "enable"))
        self.env = config.get("env", "id")
        self.env_meta = MetadataProvider().get_metadata()
        self.metrics_reporter = None
        self.metrics_registry = None
        if metrics_obj:
            self.metrics_registry = metrics_obj.metrics_registry

        self.period = self.config.getint("identifier_reporter", "period")

    def _initialize_registry(self):
        self.metrics_registry = PrometheusRegistry(batch_job=True)
        _logger.warning(f"IMPORTANT: Will get reporter {self.metrics_registry.metrics}!")
        self.metrics_reporter = get_reporter(
            self.metrics_registry,
            self.env,
            self.env_meta.task_id,
            "org_stats",
            self.env_meta.image.rsplit(":", maxsplit=1)[-1],
            chunk_size=20000,
        )
        _logger.warning(f"IMPORTANT Reporter {self.metrics_reporter}")

    def report_identifier_metrics(self):
        if not self.metrics_registry:
            self._initialize_registry()
        mode = self.config.get("service", "mode")
        if mode == "LC":
            identifier_column = "org_key"
        elif mode == "NTA":
            identifier_column = "subnet_id"
        else:
            identifier_column = None

        metric_list = self.config.get("identifier_reporter", "metrics").split(",")
        metric_stat_list = []
        _logger.warning(f"IMPORTANT Metrics list: {metric_list}")
        for metric in metric_list:
            metric_name, metric_summarization = metric.split("-")
            if self.mongodb_enabled:
                # transform to mongodb syntax
                metric_summarization = f"${metric_summarization.lower()}"
            metric_stat_list.append((metric_name, metric_summarization))

        metrics_reader_class = (
            MongoMetrics if self.mongodb_enabled else TimeStreamMetrics
        )
        # TimeStreamMetrics/MongoMetrics expects period in Seconds, multiply by 60
        metrics_reader = metrics_reader_class(
            1, self.config, identifier_column, metric_period=self.period * 60
        )
        current_time = time.time()

        rounded_time = self._round_time(current_time)
        end_time = datetime.utcfromtimestamp(rounded_time) - timedelta(microseconds=1)
        start_time = datetime.utcfromtimestamp(rounded_time) - timedelta(
            minutes=self.period
        )
        period_series_results = metrics_reader.get_recent_metrics_series_data(
            metric_stat_list, 1, start_time, end_time
        )
        _logger.warning(f"IMPORTANT: metric: {metric_stat_list}")
        _logger.warning(f"IMPORTANT current: {period_series_results}")
        for metric in metric_stat_list:
            for org_key in period_series_results[metric[0]]:
                if start_time in period_series_results[metric[0]][org_key][0]:
                    _logger.debug(
                        f"Adding {metric[0]}:{org_key}:{period_series_results[metric[0]][org_key][0][start_time]}"
                    )
                    _logger.warning(f"IMPORTANT will create a metric!!")
                    self.metrics_registry.gauge(
                        f"{metric[0]}", extra_labels={"org_key": org_key}
                    ).set_value(
                        period_series_results[metric[0]][org_key][0][start_time]
                    )
                else:
                    _logger.warning(
                        f"Missing Metric Data for time {start_time}: Org_key: {org_key} Metric: {metric[0]} "
                        f"DB return: {period_series_results[metric[0]][org_key][0]}"
                    )
        utc_end_time = end_time.replace(tzinfo=timezone.utc)
        _logger.warning(f"IMPORTANT: {self.metrics_reporter}")
        if self.metrics_reporter:
            _logger.debug("Flushing Metrics")
            _logger.warning(f"IMPORTANT: {len(self.metrics_registry.metrics)}")
            self.metrics_reporter.report_now(
                self.metrics_registry, timestamp=round(utc_end_time.timestamp())
            )
            # asyncio.run(reporter)
            _logger.debug("Finish Flushing Metrics")
        else:
            _logger.warning("No Metrics Reporter Configured")

    def _round_time(self, t):
        # round down to the most recent period end that lines up to an even partial hour.
        # Period should be evenly divisible into 60
        # Example with a 15 minute period at 4:35 round to 4:30
        return math.floor(t / (self.period * 60)) * (self.period * 60)

    def schedule_run(self):
        self.report_identifier_metrics()
        schedule.every(self.period).minutes.do(self.report_identifier_metrics)

        while True:
            schedule.run_pending()
            time.sleep(1)

    def run_once(self):
        self.report_identifier_metrics()


# This isn't normally used instead we kick off the above class with a call from mapperv2.py
# In a forked multi-processing interpreter and the values set below are set there,
# but this is still useful for running locally without the rest of org-mapper so leaving it in.
def main():
    clear_logging_handlers()
    if str_to_bool(os.environ.get("PSCR_LOGGING", "true")):
        setup_logging(
            level=os.environ.get("LOG_LEVEL", "INFO"),
            svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
            env_id=os.environ.get("ENV_ID", "UNKNOWN"),
            svc_id="TODO",
        )
    else:
        logging.basicConfig(level=os.environ.get("LOG_LEVEL", "DEBUG"))
    run_once = str_to_bool(os.environ.get("RUN_ONCE", "false"))
    org_stats_reporter = IdentifierStatsReporter(Config.create())
    if run_once:
        _logger.info("Starting org stats reporter in Run Once Mode")
        org_stats_reporter.run_once()
    else:
        _logger.info("Starting org stats reporter in Schedule Mode")
        org_stats_reporter.schedule_run()


if __name__ == "__main__":
    main()
