import json
import logging
from typing import Optional

from google.cloud import storage
from google.cloud.exceptions import NotFound

_logger = logging.getLogger(__name__)


class StorageManager:
    storage_client = storage.Client()

    @staticmethod
    def get_validated_overrides(
        manual_overrides: str, manual_overrides_backup: str, storage_bucket: str
    ) -> Optional[dict]:
        def _restore_backup() -> Optional[dict]:
            storage_backup_data = StorageManager.get_json_object(
                manual_overrides_backup, storage_bucket
            )
            if not StorageManager._validate_json_object(storage_backup_data):
                _logger.error(
                    f"Neither {manual_overrides} nor {manual_overrides_backup} are valid."
                )
                return None

            _logger.warning(
                f"Restoring override file {manual_overrides} from {manual_overrides_backup}"
            )
            StorageManager.upload_json_object(
                storage_backup_data, manual_overrides, storage_bucket
            )
            return storage_backup_data

        # noinspection PyBroadException
        try:
            storage_data = StorageManager.get_json_object(
                manual_overrides, storage_bucket
            )
            if StorageManager._validate_json_object(storage_data):
                StorageManager.upload_json_object(
                    storage_data, manual_overrides_backup, storage_bucket
                )
            else:
                storage_data = _restore_backup()
        except Exception:
            _logger.error(f"Something went wrong reading {manual_overrides}")
            storage_data = _restore_backup()
        return storage_data

    @staticmethod
    def get_validated_xdre_replays(
        manual_overrides: str, manual_overrides_backup: str, storage_bucket: str
    ) -> Optional[dict]:
        def _restore_backup() -> Optional[dict]:
            storage_backup_data = StorageManager.get_json_object(
                manual_overrides_backup, storage_bucket
            )
            if not StorageManager._validate_json_object_xdre(storage_backup_data):
                _logger.error(
                    f"Neither {manual_overrides} nor {manual_overrides_backup} are valid."
                )
                return None

            _logger.warning(
                f"Restoring override file {manual_overrides} from {manual_overrides_backup}"
            )
            StorageManager.upload_json_object(
                storage_backup_data, manual_overrides, storage_bucket
            )
            return storage_backup_data

        # noinspection PyBroadException
        try:
            storage_data = StorageManager.get_json_object(
                manual_overrides, storage_bucket
            )
            if StorageManager._validate_json_object_xdre(storage_data):
                StorageManager.upload_json_object(
                    storage_data, manual_overrides_backup, storage_bucket
                )
            else:
                storage_data = _restore_backup()
        except Exception:
            _logger.error(f"Something went wrong reading {manual_overrides}")
            storage_data = _restore_backup()
        return storage_data

    @staticmethod
    def get_json_object(file_path, storage_bucket):
        try:
            bucket = StorageManager.storage_client.get_bucket(storage_bucket)
            blob = bucket.get_blob(file_path)
            storage_data = None
            if blob:
                storage_obj = blob.download_as_bytes(checksum="crc32c")
                storage_data = storage_obj.decode("utf-8")
        except NotFound:
            storage_data = None
        except Exception as e:
            _logger.error(f"Failed to Download Object from GCS due to Error: {e}")
            raise e
        if not storage_data:
            return None

        return json.loads(storage_data)

    @staticmethod
    def upload_json_object(file_obj, file_path, storage_bucket):
        try:
            bucket = StorageManager.storage_client.get_bucket(storage_bucket)
            blob = bucket.blob(file_path)
            blob.upload_from_string(json.dumps(file_obj, indent=2))
        except Exception as e:
            _logger.exception(f"Failed to Upload Object to GCS due to Error {e}")
            raise e

    @staticmethod
    def _validate_json_object(data: dict) -> bool:
        valid = True
        if data is None:
            return False
        for override_s, override_i in data.items():
            if not isinstance(override_s, str):
                _logger.error(
                    f"Invalid value for {override_s}: {override_i}, {override_s} must be a string not {type(override_s)}."
                )
                valid = False
            if not isinstance(override_i, int):
                _logger.error(
                    f"Invalid value for {override_s}: {override_i}, {override_i} must be a integer not {type(override_i)}."
                )
                valid = False
        return valid

    @staticmethod
    def _validate_json_object_xdre(data: dict) -> bool:
        valid = True
        if data is None:
            return False
        for override_s, override_obj in data.items():
            if not isinstance(override_s, str):
                _logger.error(
                    f"Invalid value for {override_s}: {override_obj}, {override_s} must be a string not {type(override_s)}."
                )
                valid = False
            if not isinstance(override_obj, object):
                _logger.error(
                    f"Invalid value for {override_s}: {override_obj}, {override_obj} must be a object not {type(override_obj)}."
                )
                valid = False
        return valid
