import time
import logging
from math import ceil
from org_mapper.storage_manager import StorageManager
from org_mapper.utils import is_holiday_period

_logger = logging.getLogger(__name__)


class Transformer:
    def __init__(self, period_back_metric_points, metric_backend, config):
        self.env_id = config.get("env", "id")
        self.period_back_metric_points = period_back_metric_points
        self.metric_period = config.getint("cloudwatch_metrics", "metric_period")
        self.metrics = None
        self.metric_parallelism = config.getint("cloudwatch_metrics", "parallelism")
        self.top_x_points = config.getint("metrics", "top_x_points")
        self.last_x_points = config.getint("metrics", "last_x_points")
        self.metric_backend = metric_backend
        self.metric_stats_list = None
        self.manual_override_key = None
        self.manual_override_backup_key = None
        self.manual_override_priority_key = None
        self.manual_override_priority_backup_key = None
        self.bandwidth_units_per_partition = None
        self.last_hour_higher_factor = None
        self.limit_bandwidth_double = None
        self.limit_bandwidth_reduction = None
        self.holiday_check = None
        self.bandwidth_reduction_factor = None
        self.minimum_update_time_threshold_hours = None
        self.minimum_reduction_factor = None
        self.minimum_increase_factor = None
        self.admin_bucket = config.get("infra", "admin_bucket")
        self.config = config
        self.effective_traffic_config_key = None
        self.doc_builder_override_key = None
        self.effective_traffic = {}
        self.manual_overrides = None
        self.manual_overrides_priority = None
        self.xdre_replay_enabled = config.getboolean(
            "shard_management", "xdre_replay_enabled"
        )
        self.xdre_default_effective_endpoints = config.getint(
            "shard_management", "xdre_default_effective_endpoints"
        )
        self.xdre_replay_key = config.get("shard_management", "xdre_replay_key")
        self.xdre_replay_key_backup = config.get(
            "shard_management", "xdre_replay_key_backup"
        )
        self.max_shard_util = config.getfloat("shard_management", "max_shard_util")
        self.partition_keys = {}
        self.xdre_replay_orgs = {}
        self.partition_metadata = {}
        self.default_partition_priority = config.getint(
            "shard_management", "default_partition_priority"
        )
        self.xdre_partition_priority = config.getint(
            "shard_management", "xdre_partition_priority"
        )
        self.partition_key_stats = {"partial": 0, "full": 0}
        self.metric_reader = None

    def convert_bandwidth_shard_map(self):
        # returns partition data (partition keys and partition meta-data) for shard manager
        raise NotImplementedError

    def flush_docbuilder_overrides(self):
        """Round up partition counts and save docbuilder file"""
        _logger.info("Saving Doc Builder Override to Storage bucket")
        doc_builder_overrides = {}
        for identifier in self.effective_traffic:
            doc_builder_overrides[identifier] = ceil(
                self._get_id_effective_traffic_value(identifier, get_override=True)
            )
            if doc_builder_overrides[identifier] < 1:
                # edge case where no traffic for org
                doc_builder_overrides[identifier] = 1
        StorageManager.upload_json_object(
            doc_builder_overrides, self.doc_builder_override_key, self.admin_bucket
        )

    def _get_config(self):
        _logger.info("Loading Scaler Config from Storage bucket")
        effective_traffic = StorageManager.get_json_object(
            self.effective_traffic_config_key, self.admin_bucket
        )
        if not effective_traffic:
            effective_traffic = {}
        self.effective_traffic = effective_traffic

    def _save_config(self):
        _logger.info("Saving Scaler Config to Storage bucket")
        StorageManager.upload_json_object(
            self.effective_traffic, self.effective_traffic_config_key, self.admin_bucket
        )

    def get_total_effective_traffic(self):
        total_effective_traffic = 0
        for eff_traffic in self.effective_traffic:
            total_effective_traffic += self.effective_traffic[eff_traffic]["Value"]
        return total_effective_traffic

    def _remove_ids(self, remove_list):
        for identifier in remove_list:
            _logger.info(f"Removing {identifier} from the org mapper")
            del self.effective_traffic[identifier]

    def _get_manual_overrides(self):
        manual_overrides = StorageManager.get_validated_overrides(
            manual_overrides=self.manual_override_key,
            manual_overrides_backup=self.manual_override_backup_key,
            storage_bucket=self.admin_bucket,
        )
        if not manual_overrides:
            manual_overrides = {}
        return manual_overrides

    def _get_org_priority(self, identifier):
        if self.manual_overrides_priority is None:
            self.manual_overrides_priority = self._get_manual_override_priority()

        priority = self.default_partition_priority
        if self.xdre_replay_enabled and self._is_xdre_replay_started(identifier):
            priority = self.xdre_partition_priority

        return self.manual_overrides_priority.get(identifier, priority)

    def _is_xdre_replay_started(self, identifier):
        if self.xdre_replay_orgs is None:
            self.xdre_replay_orgs = self._get_xdre_replay_orgs()
        org = identifier.split(":")[0]
        return (
            org in self.xdre_replay_orgs
            and "status" in self.xdre_replay_orgs[org]
            and self.xdre_replay_orgs[org]["status"] == "STARTED"
        )

    def _get_xdre_replay_effective_endpoints(self, identifier):
        if self.xdre_replay_orgs is None:
            self.xdre_replay_orgs = self._get_xdre_replay_orgs()
        org = identifier.split(":")[0]
        if org in self.xdre_replay_orgs and "size_in_mb" in self.xdre_replay_orgs[org]:
            # 20 hours load time with throughput per second (/60 minutes /60 seconds)
            # single mb is 1024 effective endpoints, the same as one shard
            # min 1 shard to separate shard from other orgs
            # max 500 shards for a single org
            shardsNeeded = self.xdre_replay_orgs[org]["size_in_mb"] / 20 / 60 / 60
            return min(500, max(shardsNeeded, 1)) * 1024
        return self.xdre_default_effective_endpoints

    def _get_manual_override_priority(self):
        manual_overrides_priority = StorageManager.get_validated_overrides(
            manual_overrides=self.manual_override_priority_key,
            manual_overrides_backup=self.manual_override_priority_backup_key,
            storage_bucket=self.admin_bucket,
        )
        if not manual_overrides_priority:
            manual_overrides_priority = {}
        return manual_overrides_priority

    def _get_xdre_replay_orgs(self):
        xdre_replay_orgs = StorageManager.get_validated_xdre_replays(
            manual_overrides=self.xdre_replay_key,
            manual_overrides_backup=self.xdre_replay_key_backup,
            storage_bucket=self.admin_bucket,
        )
        if not xdre_replay_orgs:
            xdre_replay_orgs = {}
        return xdre_replay_orgs

    def _get_id_effective_traffic_value(self, identifier, get_override=False):
        if identifier in self.effective_traffic:
            ee = self.effective_traffic[identifier]["Value"]
        else:
            _logger.debug(
                f"Didn't find effective traffic value for {identifier} setting to 0"
            )
            ee = 0
        if get_override:
            if self.manual_overrides is None:
                self.manual_overrides = self._get_manual_overrides()
            if self.xdre_replay_enabled and self._is_xdre_replay_started(identifier):
                xdre_ee = self._get_xdre_replay_effective_endpoints(identifier)
                _logger.info(
                    f"XDRE effective endpoints for Org: {identifier} is: {xdre_ee}, and current value is {ee}"
                )
                ee = max(ee, xdre_ee)

            if identifier in self.manual_overrides:
                if ee < self.manual_overrides[identifier]:
                    _logger.info(
                        f"Manual override for Id: {identifier} "
                        f"using manual value: {self.manual_overrides[identifier]}, "
                        f"Overridden value: {ee}"
                    )
                    return self.manual_overrides[identifier]
                else:
                    _logger.info(
                        f"Manual override for Org: {identifier} but calc value is greater using: {ee}, "
                        f"Manual override value: {self.manual_overrides[identifier]}"
                    )
        return ee

    def _get_last_update_time(self, identifier):
        if identifier in self.effective_traffic:
            return self.effective_traffic[identifier]["Timestamp"]
        else:
            return None

    def _get_identifiers_to_delete(self, ids):
        remove_list = []
        for identifier in self.effective_traffic:
            if identifier not in ids:
                remove_list.append(identifier)
        return remove_list

    def _update_id_effective_traffic(self, identifier, effective_traffic_value):
        if identifier not in self.effective_traffic.keys():
            _logger.info(f"Adding {identifier} to the org mapper")
        self.effective_traffic[identifier] = {
            "Value": effective_traffic_value,
            "Timestamp": time.time(),
        }

    def _use_last_hour(self, result, last_hour_result, old_value):
        if not old_value:
            old_value = 0
        #  Don't decrease if last hour is below old value but above look back longer calculated value
        if result < last_hour_result < old_value:
            return old_value
        elif last_hour_result >= result * self.last_hour_higher_factor:
            return last_hour_result
        else:
            return result

    def _calc_number_partition_keys(self, effective_traffic):
        return 1 + int(effective_traffic * 1.0 / self.bandwidth_units_per_partition)

    def _get_shard_percent(self, effective_traffic):
        return effective_traffic / self.bandwidth_units_per_partition

    def _update_partition_keys(self):
        for identifier in self.effective_traffic.keys():
            effective_bandwidth = self._get_id_effective_traffic_value(
                identifier, get_override=True
            )
            num_partition_keys = self._calc_number_partition_keys(effective_bandwidth)
            i = 0
            while i < num_partition_keys:
                key_hash = self._generate_kinesis_hash(identifier, i)
                if effective_bandwidth < self.bandwidth_units_per_partition:
                    self.partition_keys[key_hash] = self._get_shard_percent(
                        effective_bandwidth
                    )
                    self.partition_metadata[key_hash] = {
                        "name": f"{identifier}:{i}",
                        "percent": self._get_shard_percent(effective_bandwidth),
                    }
                    self.partition_key_stats["partial"] += 1
                else:
                    self.partition_keys[key_hash] = self.max_shard_util
                    self.partition_metadata[key_hash] = {
                        "name": f"{identifier}:{i}",
                        "percent": 1,
                    }
                    self.partition_key_stats["full"] += 1

                self.partition_metadata[key_hash]["priority"] = self._get_org_priority(
                    identifier
                )
                i += 1

    @staticmethod
    def _generate_kinesis_hash(org, partitionid):
        # Implement in subclass
        raise NotImplementedError

    def limit_bandwidth_movement(self, effective_bandwidth, old_value, identifier):
        if self.limit_bandwidth_double:
            #  Limit to doubling previous value
            if old_value is not None and old_value != 0:
                if effective_bandwidth > (old_value * 2):
                    _logger.debug(
                        f"Indicated effective bandwidth more than double, only doubling for {identifier}"
                    )
                    effective_bandwidth = old_value * 2
        if self.limit_bandwidth_reduction:
            if old_value is not None:
                #  Limit reduction to configurable factor.
                compare_value = int((old_value * (1 - self.bandwidth_reduction_factor)))
                if effective_bandwidth < compare_value:
                    _logger.debug(
                        f"Indicated effective bandwidth decrease more than "
                        f"factor({self.bandwidth_reduction_factor}, only reducing to "
                        f"{compare_value} for {identifier}"
                    )
                    effective_bandwidth = compare_value
        if self.holiday_check and is_holiday_period(self.config):
            if old_value is not None and effective_bandwidth < old_value:
                _logger.info(
                    f"Holiday Period Check: old effective bandwidth {old_value} is greater than current effective "
                    f"bandwidth {effective_bandwidth}, using old effective bandwidth for the org {identifier}"
                )
                effective_bandwidth = old_value

        return effective_bandwidth

    def update_bandwidth_rules(
        self, new_value, last_update_time, identifier, old_value
    ):
        # Don't update if last update time is less than hours ago
        if last_update_time:
            if last_update_time > (
                time.time() - (3600 * self.minimum_update_time_threshold_hours)
            ):
                _logger.info(
                    f"Last update of effective bandwidth for {identifier} was too recent, "
                    f"needs to be at least {self.minimum_update_time_threshold_hours} ago."
                )
                return False
        # Don't update if we are within configurable value of current value
        if old_value:
            if (
                old_value
                <= new_value
                < (old_value * self.minimum_increase_factor) + old_value
            ):
                _logger.debug(
                    f"Effective Bandwidth Increase for {identifier} within "
                    f"{self.minimum_increase_factor}, not updating"
                )
                return False
            if (
                old_value
                > new_value
                > old_value - (old_value * self.minimum_reduction_factor)
            ):
                _logger.debug(
                    f"Effective Bandwidth Decrease for {identifier} within "
                    f"{self.minimum_reduction_factor}, not updating"
                )
                return False
        return True
