import datetime
import time
import logging
import os
from org_mapper.app_metrics import Metrics
from pscr_common.config import Config as _Config

_logger = logging.getLogger(__name__)


def is_holiday_period(config):
    """
    This function returns True if its Holiday Period
    :param config:
    :return:
    """
    curr_date = datetime.datetime.now()
    curr_year = curr_date.year

    # start date and end date are accepted in month/day format (default dates are set to 12/20 to 1/8)
    # start date is for current year and end date is always for next year
    # ( ex: for year 2021 date range (start date) 12/20 to (end date) 1/18 represents 2021-12-20 to 2022-01-08)

    try:
        start_date_month = int(
            config.get("metrics", "holiday_period_start_date").split("/")[0]
        )
        start_date_day = int(
            config.get("metrics", "holiday_period_start_date").split("/")[1]
        )

        date_check_1_start = datetime.datetime(
            curr_year, start_date_month, start_date_day
        )
        date_check_1_end = datetime.datetime(curr_year, 12, 31)

        end_date_month = int(
            config.get("metrics", "holiday_period_end_date").split("/")[0]
        )
        end_date_day = int(
            config.get("metrics", "holiday_period_end_date").split("/")[1]
        )

        date_check_2_start = datetime.datetime(curr_year, 1, 1)
        date_check_2_end = datetime.datetime(curr_year, end_date_month, end_date_day)

        # if start date is not in Dec and End date is not in Jan unexpected things might happen
        if start_date_month != 12 or end_date_month != 1:
            _logger.warning(
                f"Start Date : {start_date_month}/{start_date_day} and/or End Date :{end_date_month}/{end_date_day} "
                f"are set different from default date range DEC-JAN.Unexpected things might happen"
            )

    except ValueError as e:
        _logger.error(e)
        _logger.warning(f"Setting Default Holiday Start and End Dates due to error {e}")
        # Default Start Date and End Date are set from DEC 20th to JAN 8th
        date_check_1_start = datetime.datetime(curr_year, 12, 20)
        date_check_1_end = datetime.datetime(curr_date.year, 12, 31)
        date_check_2_start = datetime.datetime(curr_year, 1, 1)
        date_check_2_end = datetime.datetime(curr_date.year, 1, 8)

    if (date_check_1_start.date() <= curr_date.date() <= date_check_1_end.date()) or (
        date_check_2_start.date() <= curr_date.date() <= date_check_2_end.date()
    ):
        return True

    return False


def output_stats(transformer, shard_manager):
    #  Divide by 2 because we have default and long term shard keys for each org
    _logger.info(
        f"We have {len(transformer.effective_traffic) / 2} identifiers in this environment"
    )

    for stat in transformer.partition_key_stats:
        _logger.info(
            f"We have {transformer.partition_key_stats[stat]} {stat} partition keys"
        )

    splits = "Split Shards: \n"
    for shard in shard_manager.shard_splits:
        splits += f"{shard[0]}\n"
    _logger.info(splits)

    merges = "Merged Shards: \n"
    for shard in shard_manager.shard_merges:
        merges += f"{shard[0]}:{shard[1]}\n"
    _logger.info(merges)

    _logger.info(
        f"Started with {len(shard_manager.initial_active_shards)} shards ended with "
        f"{len(shard_manager.get_active_shards())} shards, "
        f"Split {len(shard_manager.shard_splits)} shards, "
        f"Merged {len(shard_manager.shard_merges)} shards"
    )

    shard_size_map = {0.25: 0, 0.5: 0, 0.75: 0, 1.0: 0, 1.25: 0, 1.50: 0}
    for shard in shard_manager.new_shards_map:
        shard_size = shard[2]
        if shard_size < 0.25:
            shard_size_map[0.25] += 1
        elif shard_size < 0.5:
            shard_size_map[0.5] += 1
        elif shard_size < 0.75:
            shard_size_map[0.75] += 1
        elif shard_size < 1:
            shard_size_map[1.0] += 1
        elif shard_size < 1.25:
            shard_size_map[1.25] += 1
        else:
            shard_size_map[1.50] += 1

    shard_percent = "Distribution of Shards not belonging to large orgs\n"
    shard_percent += f"We have {shard_size_map[.25]} shards 0 to 25 percent full\n"
    shard_percent += f"We have {shard_size_map[.5]} shards 25 to 50 percent full\n"
    shard_percent += f"We have {shard_size_map[.75]} shards 50 to 75 percent full\n"
    shard_percent += f"We have {shard_size_map[1.0]} shards 75 to 100 percent full\n"
    shard_percent += f"We have {shard_size_map[1.25]} shards 100 to 125 percent full\n"
    shard_percent += (
        f"We have {shard_size_map[1.50] - transformer.partition_key_stats['full']} "
        f"shards 125 to 150 percent full\n"
    )

    _logger.info(shard_percent)

    _logger.info(f"Top 10 effective endpoints")
    i = 1

    top_ten = ""

    for effective_endpoint in sorted(
        transformer.effective_traffic,
        key=lambda x: (transformer.effective_traffic[x]["Value"]),
        reverse=True,
    ):
        if i <= 10:
            top_ten += f"{effective_endpoint}:{transformer.effective_traffic[effective_endpoint]['Value']}\n"
            i += 1

    _logger.info(top_ten)

    _logger.info(
        f"Total Effective Endpoints: {transformer.get_total_effective_traffic()}"
    )

    _logger.debug(shard_manager.new_shards_map)
    _logger.debug(transformer.effective_traffic)


def str_to_bool(value):
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    raise Exception("Invalid boolean sring passed as env var")


def log_and_sleep(message, retries=0, base_sleep=0.1, max_sleep=120):
    """
    Utility function to use for exponential backoff when retrying
    """
    metrics_obj = Metrics()
    metrics_obj.start_reporter()
    metrics_obj.metrics_registry.counter(
        f"org_mapper.error.{message.replace(' ', '_')}"
    )
    sleep_time = min((2**retries) * base_sleep, max_sleep)
    if sleep_time < max_sleep:
        _logger.debug(message)
    else:
        _logger.error(message)
        raise

    time.sleep(sleep_time)
    return message


class Config(_Config):
    CONFIG_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.ini")
